using Gtk;
using GLib;

namespace Boilerplate {

    public class Application : Gtk.Application {

        private const GLib.OptionEntry[] option_entries = {
            { "debug", 'd', 0, OptionArg.NONE, null, "Run in debug mode and show extra output", null },
            { "version", 'v', 0, OptionArg.NONE, null, "Print release version and exit", null },
            { "disable-csd", 'c', 0, OptionArg.INT, null, "Disable Client Side Decorations (CSD)", null },
            { null }
        };

        private const GLib.ActionEntry[] action_entries = {
            { "preferences", on_preferences },
            { "shortcuts", on_shortcuts },
            { "help", on_help },
            { "about", on_about },
            { "quit", on_quit },
        };

        public GLib.Settings settings;
        public Window window;
        public bool csd = true;

        /**
         * (1) Start here.
         * Constructor for application, starting base Gtk.Application object is
         * done here as well as setting application name, getting application 
         * settings and anything that needs to be done before application can
         * continue.
         */
        public
        Application (args) {
            GLib.Object (application_id: "io.github.julianrichen.boilerplate",
                         flags: GLib.ApplicationFlags.FLAGS_NONE);

            GLib.debug ("Starting Application");

            stdout.printf ("From Config.vala: %s", Config.ID);
            
            Environment.set_application_name ("Boilerplate");

            this.add_main_option_entries (option_entries);

            /**
             * Settings look-up and loader. First we check relative to the binaries
             * execution location. Ex. If our executable is:
             * ./bin/exec
             *  
             * We look recursivly in:
             * ./share/... (which is ./share/glib-2.0/schemas)
             * 
             * Otherwise we look in the default XDG_DATA_DIRS, which are:
             * /usr/local/share/glib-2.0/schemas
             * /usr/share/glib-2.0/schemas
             */
            GLib.debug ("Looking-up gschema settings file");
            string settings_dir = Path.get_dirname (args[0]);
            SettingsSchemaSource sss = new Gio.SettingsSchemaSource.from_directory (settings_dir, null, false);
            SettingsSchema schema = sss.lookup ("io.github.julianrichen.boilerplate", true);
            
            if (sss.lookup != null) {
                GLib.debug ("Loading relative gschema settings file from: %s", settings_dir);
                Settings settings = new Gio.Settings.full (schema, null, null);
            } else {
                Glib.warning ("Could not locate relative settings file, trying XDG_DATA_DIRS default");
                
                SettingsSchemaSource sss = new Gio.SettingsSchemaSource.get_default()
                SettingsSchema schema = sss.lookup ("io.github.julianrichen.boilerplate", true)

                if (sss.lookup != null) {
                    GLib.debug ("Loading default gschema settings file");
                    Settings settings = new Gio.Settings ("io.github.julianrichen.boilerplate");
                } else {
                    GLib.critical ("Could not load any gschema settings file, ABORT!");
                    return 1;
                }
            }
        }

        /**
         * (2) Parse and handle command line arguemnts first.
         * Handle all command line options defined in 
         * GLib.OptionEntry[] option_entries.
         */
        protected override int
        handle_local_options (GLib.VariantDict options) {
            GLib.debug ("Handling local options");
            
            if (options.contains("version")) {
                //print ("%s %s\n", Environment.get_application_name (), Config.VERSION);
                return 0;
            }

            if (options.contains("debug")) {
                // print("Starting debug...");
            }

            if (options.contains("disable-csd")) {
                this.csd = false;
            }

            return -1;
        }

        /**
         * (3) Start the application up.
         * Start basic operations for application. Ex. Creating application menu
         * or starting a database/daemon.
         */
        protected override void
        startup () {
            GLib.debug ("Starting application");
            
            base.startup ();
            
            GLib.debug ("Building applicationb menu");
            this.add_action_entries (action_entries, this);

            var builder = new Gtk.Builder ();

            try {
                builder.add_from_file ("menu.ui");
            } catch (Error e) {
                GLib.error ("Unable to load file: %s", e.message);
            }

            this.app_menu = builder.get_object ("app-menu") as MenuModel;
        }

        /**
         * (4) Activate window and show all.
         * Creating or presenting the main window is done here.
         */
        protected override void
        activate () {
            GLib.debug ("Activating application");
            
            base.activate();
            
            if (this.get_windows () != null) {
                GLib.debug ("Window already created, presenting current window");
                this.window.present ();
            } else {
                GLib.debug ("Creating new main window");
                this.window = new Window (this);

                this.window.destroy.connect (() => {
                    this.window = null;
                });
                
                this.add_window(this.window);

                this.window.show_all();
            }
        }

        /**
         * (5) Shutdown application.
         * Any operatings that should be done before the application quits is
         * done here. Ex. Saving files or updating a database.
         */
        protected override void
        shutdown () {
            GLib.debug ("Application shutting down");
            base.shutdown();
        }

        private void
        do_preferences (GLib.SimpleAction action, GLib.Variant? parameter) {
            GLib.debug ("Opening preferences dialog");
            
            stdout.printf ("Preferences");
            
            if (this.preferences_window != null) {
                GLib.debug ("[Preferences] Window already created, presenting current window");
                this.preferences_window.present ();
            } else {
                GLib.debug ("[Preferences] Creating new dialog");
                this.preferences_window = new PreferencesWindow ();

                this.preferences_window.destroy.connect (() => {
                    this.preferences_window = null;
                });

                this.preferences_window.show_all();
            }
        }

        private void
        do_shortcuts (GLib.SimpleAction action, GLib.Variant? parameter) {
            GLib.debug ("Opening keyboard shortcuts dialog");
            
            stdout.printf ("Keyboard Shortcuts");
        }

        private void
        do_help (GLib.SimpleAction action, GLib.Variant? parameter) {
            GLib.debug ("Opening help dialog");
            
            try {
                var screen = this.window.get_screen ();
                Gtk.show_uri (screen,
                              "help:boilerplate",
                              Gtk.get_current_event_time ());
            } catch (GLib.Error e) {
                GLib.warning ("Failed to display help: %s", e.message);
            }
        }

        private void
        do_about (GLib.SimpleAction action, GLib.Variant? parameter) {
            GLib.debug ("Opening about dialog");
            string[] authors = {
                "Julian Richen <julian@richen.io>",
                null
            };

            string[] artists = {
                "Julian Richen <julian@richen.io>",
                null
            };

            string[] documenters = {
                "Julian Richen <julian@richen.io>",
                null
            };

            Gtk.show_about_dialog (
                this.window,
                "program-name", "Boilerplate for Vala & GTK 3",
                "authors", authors,
                "artists", artists,
                "documenters", documenters,
                "version", "0.1.0",
                "license-type", Gtk.License.GPL_3_0,
                "logo-icon-name", "io.github.julianrichen.boilerplate",
                "website", "https://github.com/julianrichen/boilerplate-for-vala-gtk3"
            );
        }

        private void
        do_quit (GLib.SimpleAction action, GLib.Variant? parameter) {
            GLib.debug ("Application quitting");
            base.quit ();
        }

    }

}
